/* Portions (c) Copyright 1996 Novell, Inc. All rights reserved. */
/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Username handling
   Copyright (C) Andrew Tridgell 1992-1995
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "includes.h"
#include "loadparm.h"

/****************************************************************************
get a users home directory. tries as-is then lower case
****************************************************************************/
char *get_home_dir(char *user)
{
  struct passwd *pass;

  pass = Get_Pwnam(user,False);

  if (!pass) return(NULL);
  return(pass->pw_dir);      
}

#ifdef NETWARE
struct passwd*
Get_Pwnam(char* user,BOOL allow_change)
{
  extern int guestConnection;
  struct passwd* pwd = 0;
  Buf_T* Filter = 0;
  Buf_T* Names = 0;
  Buf_T* Result = 0;
  int32 iteration = -1;
  Filter_Cursor_T* cursor = 0;
  extern NWDSContextHandle nwdsContext;
  
  /* allocate everything first */
  if (NWDSAllocBuf(DEFAULT_MESSAGE_LEN, &Names) < 0)
    goto error1;
  if (NWDSAllocBuf(DEFAULT_MESSAGE_LEN, &Filter) < 0)
    goto error2;
  if (NWDSAllocBuf(DEFAULT_MESSAGE_LEN, &Result) < 0)
    goto error3;
  
  /* initialize input buffers */
  if (NWDSInitBuf(nwdsContext, DSV_SEARCH, Names) < 0)
    goto release;
  if (NWDSInitBuf(nwdsContext, DSV_SEARCH_FILTER, Filter) < 0)
    goto release;

  /* allocate and load the search filter expression tree */
  if (NWDSAllocFilter(&cursor) < 0)
    goto release;
  if (NWDSAddFilterToken(cursor,
			 FTOK_ANAME,
			 "Object Class",
			 SYN_CLASS_NAME) < 0)
    goto error4;
  if (NWDSAddFilterToken(cursor, FTOK_EQ, 0, 0) < 0)
    goto error4;
  if (NWDSAddFilterToken(cursor,
			 FTOK_AVAL,
			 "User",
			 SYN_CLASS_NAME) < 0)
    goto error4;
  if (NWDSAddFilterToken(cursor, FTOK_END, 0, 0) < 0) {
  error4:
    NWDSFreeFilter(cursor, 0);
    goto release;
  }
  if (NWDSPutFilter(nwdsContext, Filter, cursor, 0) < 0)
    goto release;

  /* grab home directory along the way */
  if (NWDSPutAttrName(nwdsContext, Names, "Home Directory") < 0)
    goto release;

  do {
    uint32 nObj;    
    if (NWDSSearch(nwdsContext,		/* context */
		   user,		/* baseObjectName */
		   DS_SEARCH_ENTRY,	/* scope */
		   TRUE,		/* searchAliases */
		   Filter,		/* filter */
		   1,			/* infoType, names and values */
		   FALSE,		/* allAttrs */
		   Names,		/* attrNames */
		   &iteration,		/* iterationHandle */
		   0,			/* countObjectToSearch */
		   0,			/* countObjectSearched */
		   Result) < 0)		/* objectInfo */
      goto release;

    /* user exists! record some info */
    pwd = &passwd_ent;
    free(pwd->pw_name);
    pwd->pw_name = strdup(user);
    if (pwd->pw_name == NULL) {
      pwd = NULL;
      goto release;
    }
    pwd->pw_uid = (strequal(pwd->pw_name, lp_guestaccount(-1)) ?
		   guestConnection :
		   GetCurrentConnection());
    /* pwd->pw_gid = 0; doesn't change */
    free(pwd->pw_dir);
    pwd->pw_dir = strdup("SYS:PUBLIC");	/* they may not have a "home" */
    if (pwd->pw_dir == NULL) {
      pwd = NULL;
      goto release;
    }

    if (NWDSGetObjectCount(nwdsContext, Result, &nObj) < 0)
      goto release;
    while (nObj-- > 0) {
      uint32 nAttr;
      char buf[MAX_DN_CHARS];
      Object_Info_T objInfo;
      if (NWDSGetObjectName(nwdsContext, Result, buf, &nAttr, &objInfo) < 0)
	goto release;
      while (nAttr-- > 0) {
	uint32 nVal, synID;
	if (NWDSGetAttrName(nwdsContext, Result, buf, &nVal, &synID) < 0)
	  goto release;
	if (synID != SYN_PATH || strcmp(buf, "Home Directory") != ESUCCESS)
	  continue;
	if (NWDSGetAttrVal(nwdsContext, Result, synID, buf) < 0)
	  goto release;

	/* got everything i could ever want! */
	free(pwd->pw_dir);
	pwd->pw_dir = strdup(buf);
	goto release;
      }
    }
  } while (iteration > -1);

 release:
  NWDSFreeBuf(Result);
 error3:
  NWDSFreeBuf(Filter);
 error2:
  NWDSFreeBuf(Names);
 error1:
  return pwd;
}

#else /* NETWARE */

/*******************************************************************
map a username from a dos name to a unix name by looking in the username
map
********************************************************************/
void map_username(char *user)
{
  static int depth=0;
  static BOOL initialised=False;
  static fstring last_from,last_to;
  FILE *f;
  char *s;
  char *mapfile = lp_username_map();
  if (!*mapfile || depth) return;

  if (!*user) return;

  if (!initialised) {
    *last_from = *last_to = 0;
    initialised = True;
  }

  if (strequal(user,last_to)) return;

  if (strequal(user,last_from)) {
    DEBUG(3,("Mapped user %s to %s\n",user,last_to));
    strcpy(user,last_to);
    return;
  }
  
  f = fopen(mapfile,"r");
  if (!f) {
    DEBUG(0,("can't open username map %s\n",mapfile));
    return;
  }

  DEBUG(4,("Scanning username map %s\n",mapfile));

  depth++;

  for (; (s=fgets_slash(NULL,80,f)); free(s)) {
    char *unixname = s;
    char *dosname = strchr(unixname,'=');

    if (!dosname) continue;
    *dosname++ = 0;

    while (isspace(*unixname)) unixname++;
    if (!*unixname || strchr("#;",*unixname)) continue;

    {
      int l = strlen(unixname);
      while (l && isspace(unixname[l-1])) {
	unixname[l-1] = 0;
	l--;
      }
    }

    if (strchr(dosname,'*') || user_in_list(user,dosname)) {
      DEBUG(3,("Mapped user %s to %s\n",user,unixname));
      StrnCpy(last_from,user,sizeof(last_from)-1);
      sscanf(unixname,"%s",user);
      StrnCpy(last_to,user,sizeof(last_to)-1);
    }
  }

  fclose(f);

  depth--;
}

/****************************************************************************
internals of Get_Pwnam wrapper
****************************************************************************/
static struct passwd *_Get_Pwnam(char *s)
{
  struct passwd *ret;

  ret = getpwnam(s);
  if (ret)
    {
#ifdef GETPWANAM
      struct passwd_adjunct *pwret;
      pwret = getpwanam(s);
      if (pwret)
	{
	  free(ret->pw_passwd);
	  ret->pw_passwd = pwret->pwa_passwd;
	}
#endif

    }

  return(ret);
}


/****************************************************************************
a wrapper for getpwnam() that tries with all lower and all upper case 
if the initial name fails. Also tried with first letter capitalised
Note that this changes user!
****************************************************************************/
struct passwd *Get_Pwnam(char *user,BOOL allow_change)
{
  fstring user2;

  struct passwd *ret;  

  if (!user || !(*user))
    return(NULL);

  StrnCpy(user2,user,sizeof(user2)-1);

  if (!allow_change) {
    user = &user2[0];
  }

  map_username(user);

  ret = _Get_Pwnam(user);
  if (ret) return(ret);

  strlower(user);
  ret = _Get_Pwnam(user);
  if (ret)  return(ret);

  strupper(user);
  ret = _Get_Pwnam(user);
  if (ret) return(ret);

  /* try with first letter capitalised */
  if (strlen(user) > 1)
    strlower(user+1);  
  ret = _Get_Pwnam(user);
  if (ret) return(ret);

  if (allow_change)
    strcpy(user,user2);

  return(NULL);
}
#endif /* NETWARE */


/****************************************************************************
check if a user is in a user list
****************************************************************************/
BOOL user_in_list(char *user,char *list)
{
  pstring tok;
  char *p=list;

  while (next_token(&p,tok,LIST_SEP))
    {
      if (strequal(user,tok))
	return(True);

#ifdef NETGROUP
      if (*tok == '@')
	{
	  static char *mydomain = NULL;
	  if (mydomain == 0)
	    yp_get_default_domain(&mydomain);
	  
	  DEBUG(5,("looking for user %s of domain %s in netgroup %s\n",
		   user, mydomain, &tok[1]));
	  DEBUG(5,("innetgr is %s\n",
		   innetgr(&tok[1], (char *) 0, user, mydomain)
		   ? "TRUE" : "FALSE"));
	  
	  if (innetgr(&tok[1], (char *)0, user, mydomain))
	    return (True);
	}
#endif


#if HAVE_GETGRNAM 
      if (*tok == '@')
	{
          struct group *gptr;
          char **member;  
	  struct passwd *pass = Get_Pwnam(user,False);

	  if (pass) { 
	    gptr = getgrgid(pass->pw_gid);
	    if (gptr && strequal(gptr->gr_name,&tok[1]))
	      return(True); 
	  } 

	  gptr = (struct group *)getgrnam(&tok[1]);

	  if (gptr)
	    {
	      member = gptr->gr_mem;
	      while (member && *member)
		{
		  if (strequal(*member,user))
		    return(True);
		  member++;
		}
	    }
	}	      
#endif
    }
  return(False);
}


